-- Exercise in Eden programming: Hamming numbers
--
-- The sequence of Hamming numbers consists of all multiples
-- of 2, 3, and 5, in ascending order, without duplicates:
--     { 2^i * 3^j * 5^k | i,j,k natural numbers}
--
-- Observation (Dijkstra): 
-- 1) 1 is a Hamming number
-- 2) If h is a Hamming number, 2*h, 3*h and 5*h are also 
--    Hamming numbers. 
--
------------------------------------------------------------

import Control.Parallel.Eden hiding (merge)
import System.Environment

main :: IO ()
main = do args <- getArgs
          let n   = if null args then 0 else read (head args)
              res = take n hamming
          putStr ("Hamming numbers up to " ++ show n ++ ":\n" ++ show res)


hamming = let [h2,h3,h5] = spawn (map times [2,3,5]) (repeat hamming)
          in 1 : merge (merge h2 h3) h5

times :: Int -> Process [Int] [Int]
times n = process (map (*n))

merge :: [Int] -> [Int] -> [Int]
--merge [] bs = bs
--merge as [] = as
merge as@(a:ra) bs@(b:rb) 
    = case compare a b of
        LT -> a: merge ra bs
        GT -> b: merge as rb
        EQ -> a: merge ra rb
 
mergeP :: Process ([Int],[Int]) [Int]
mergeP = process (uncurry merge)
